# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest.mock import Mock, patch

import pytest

from fibratus.apidefs.registry import HKEY_LOCAL_MACHINE, HKEY_USERS
from fibratus.common import DotD as dd, NA
from fibratus.handle import HandleInfo, HandleType
from fibratus.kevent import KEvent
from fibratus.kevent_types import *
from fibratus.registry import HiveParser, Kcb
from fibratus.thread import ThreadRegistry, ThreadInfo


@pytest.fixture(scope='module')
def kevent_mock():
    return Mock(spec_set=KEvent)


@pytest.fixture(scope='module')
def thread_registry_mock():
    thread_registry = Mock(spec_set=ThreadRegistry)
    thread_info = ThreadInfo(896, 2916, 22, 'explorer.exe', 'C:\\Windows\\EXPLORER.exe')
    thread_info.handles.append(HandleInfo(836, 18446735964859105184, HandleType.KEY,
                               "\\REGISTRY\\USER\\S-1-5-21-2945379629-2233710143-2353048178-1000_CLASSES\\Local Settings"
                               "\\Software\Microsoft\\Windows\\Shell\\Bags\\59\\Shell\\{5C4F28B5-F869-4E84-8E60-F11DB97C5CC7}",
                               896))
    thread_registry.get_thread.return_value = thread_info
    return thread_registry


@pytest.fixture(scope='module')
def hive_parser(kevent_mock, thread_registry_mock):
    kcb1 = dd({"index": 0, "process_id": 1224, "status": 0, "key_handle": 18446735964840821928,
              "key_name": "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\MountPoints2\\"
                          "{4a33e644-94a8-11e5-a0c5-806e6f6e6963}\\",
              "thread_id": 1484, "initial_time": 24218562806})
    kcb2 = dd({'initial_time': 0, 'index': 0, 'thread_id': 620, 'status': 0, 'key_handle': 18446735964896987168,
               'key_name': '\\REGISTRY\\MACHINE\\SYSTEM\\ControlSet001\\services\\HDAudBus', 'process_id': 3820})
    kcb3 = dd({"index": 0, "process_id": 896, "status": 0, "key_handle": 18446735964812642928,
               "key_name": "\\REGISTRY\\MACHINE\\SOFTWARE\\Classes\\CLSID\\{D2D588B5-D081-11D0-99E0-00C04FC2F8EC}"
                           "\\InprocServer32", "thread_id": 2916, "initial_time": 0})
    hive_parser = HiveParser(kevent_mock, thread_registry_mock)
    hive_parser.add_kcb(kcb1)
    hive_parser.add_kcb(kcb2)
    hive_parser.add_kcb(kcb3)
    return hive_parser


class TestHiveParser():

    def test_add_kcb(self, hive_parser):
        kcb = dd({'initial_time': 0, 'index': 0, 'thread_id': 620,
                  'status': 0, 'key_handle': 18446735964879434920,
                  'key_name': '\\REGISTRY\\MACHINE\\SOFTWARE\\Microsoft\\WBEM\\WDM', 'process_id': 3820})
        hive_parser.add_kcb(kcb)
        assert kcb.key_handle in hive_parser.kcblocks

        kcblock = hive_parser.kcblocks[kcb.key_handle]
        assert isinstance(kcblock, Kcb)
        assert kcblock.key == kcb.key_name

    def test_remove_kcb(self, hive_parser):
        kcb = dd({'initial_time': 0, 'index': 0, 'thread_id': 620,
                  'status': 0, 'key_handle': 18446735964879434920,
                  'key_name': '\\REGISTRY\\MACHINE\\SOFTWARE\\Microsoft\\WBEM\\WDM', 'process_id': 3820})
        hive_parser.remove_kcb(kcb.key_handle)
        assert kcb.key_handle not in hive_parser.kcblocks

    @pytest.mark.parametrize('kevent_type', [REG_OPEN_KEY, REG_CREATE_KEY, REG_QUERY_KEY, REG_DELETE_KEY])
    def test_parse_hive_key_kevent_full_node_name(self, kevent_type, hive_parser, kevent_mock):
        regkevt = dd({"index": 0, "process_id": 1224, "status": 3221225524, "key_handle": 0,
                      "key_name": "\\Registry\\Machine\\Software\\Classes\\Applications\\Explorer.exe"
                                  "\\Drives\\C\\DefaultIcon", "thread_id": 2164, "initial_time": 24218563385})
        hive_parser.parse_hive(kevent_type, regkevt)
        kparams = kevent_mock.params

        assert kparams['hive'] == 'REGISTRY_MACHINE_SOFTWARE'
        assert kparams['key'] == 'SOFTWARE\\Classes\\Applications\\Explorer.exe\\Drives\\C\\DefaultIcon'
        assert kparams['status'] == regkevt.status
        assert kparams['pid'] == regkevt.process_id
        assert kparams['tid'] == regkevt.thread_id

    @pytest.mark.parametrize('kevent_type', [REG_OPEN_KEY, REG_CREATE_KEY, REG_QUERY_KEY, REG_DELETE_KEY])
    def test_parse_hive_key_kevent_kcb_lookup_match(self, kevent_type, hive_parser, kevent_mock):
        regkevt = dd({"index": 2, "process_id": 896, "status": 0, "key_handle": 18446735964812642928,
                      "key_name": "ThreadingModel", "thread_id": 2916, "initial_time": 24219715376})
        hive_parser.parse_hive(kevent_type, regkevt)
        kparams = kevent_mock.params

        assert kparams['hive'] == 'REGISTRY_MACHINE_SOFTWARE'
        assert kparams['key'] == 'SOFTWARE\\Classes\\CLSID\\{D2D588B5-D081-11D0-99E0-00C04FC2F8EC}' \
                                  '\\InprocServer32\\ThreadingModel'
        assert kparams['status'] == regkevt.status
        assert kparams['pid'] == regkevt.process_id
        assert kparams['tid'] == regkevt.thread_id

    def test_parse_hive_key_handles_lookup(self, hive_parser, kevent_mock, thread_registry_mock):
        regkevt = dd({"index": 0, "process_id": 896, "status": 3221225524,
                      "key_handle": 18446735964819421216,
                      "key_name": "Bags\\59\\Shell\\{5C4F28B5-F869-4E84-8E60-F11DB97C5CC7}",
                      "thread_id": 2916, "initial_time": 24219715717})
        hive_parser.parse_hive(REG_OPEN_KEY, regkevt)

        thread_registry_mock.get_thread.assert_called_with(regkevt.process_id)

        kcb = hive_parser.kcblocks[regkevt.key_handle]
        assert kcb and isinstance(kcb, Kcb)

        assert kcb.key == "\\REGISTRY\\USER\\S-1-5-21-2945379629-2233710143-2353048178-1000_CLASSES\\Local Settings" \
                          "\\Software\Microsoft\\Windows\\Shell\\Bags\\59" \
                          "\\Shell\\{5C4F28B5-F869-4E84-8E60-F11DB97C5CC7}"
        assert kevent_mock.params['hive'] == 'REGISTRY_USER_S-1-5-21-2945379629-2233710143-2353048178-1000_CLASSES'
        assert kevent_mock.params['key'] == "S-1-5-21-2945379629-2233710143-2353048178-1000_CLASSES\\Local Settings" \
                                            "\\Software\Microsoft\\Windows\\Shell\\Bags\\59" \
                                            "\\Shell\\{5C4F28B5-F869-4E84-8E60-F11DB97C5CC7}"

    def test_parse_hive_set_value_kevent_full_node_name(self, hive_parser, kevent_mock):
        regkevt = dd({"index": 0, "process_id": 1224, "status": 3221225524, "key_handle": 0,
                     "key_name": "\\Registry\\Machine\\Software\\Classes\\Applications\\Explorer.exe"
                     "\\Drives\\C\\DefaultIcon", "thread_id": 2164, "initial_time": 24218563385})
        with patch('fibratus.registry.HiveParser._query_value', return_value=('open', 'REG_SZ')) \
                as query_value_mock:
            hive_parser.parse_hive(REG_SET_VALUE, regkevt)
            kparams = kevent_mock.params
            query_value_mock.assert_called_with(HKEY_LOCAL_MACHINE,
                                                'SOFTWARE\\Classes\\Applications\\Explorer.exe\\Drives\\C',
                                                'DefaultIcon')

            assert kparams['hive'] == 'REGISTRY_MACHINE_SOFTWARE'
            assert kparams['key'] == 'SOFTWARE\\Classes\\Applications\\Explorer.exe\\Drives\\C\\DefaultIcon'
            assert kparams['status'] == regkevt.status
            assert kparams['pid'] == regkevt.process_id
            assert kparams['tid'] == regkevt.thread_id
            assert kparams['value_type'] == 'REG_SZ'
            assert kparams['value'] == 'open'

    def test_parse_hive_set_value_kevent_full_node_name(self, hive_parser, kevent_mock):
        regkevt = dd({"index": 0, "process_id": 1224, "status": 3221225524, "key_handle": 0,
                              "key_name": "\\REGISTRY\\USER\\S-1-5-21-2945379629-2233710143-2353048178-1000_CLASSES"
                              "\\Local Settings\\Software\\Microsoft\\Windows\\CurrentVersion\\TrayNotify\\IconStreams",
                              "thread_id": 2164, "initial_time": 24218563385})
        with patch('fibratus.registry.HiveParser._query_value', return_value=('0x12', 'REG_DWORD')) \
                as query_value_mock:
            hive_parser.parse_hive(REG_QUERY_VALUE, regkevt)
            kparams = kevent_mock.params
            query_value_mock.assert_called_with(HKEY_USERS,
                                                    "S-1-5-21-2945379629-2233710143-2353048178-1000_CLASSES"
                                                    "\\Local Settings\\Software\\Microsoft\\Windows\\CurrentVersion"
                                                    "\\TrayNotify",
                                                    'IconStreams')
            assert kparams['hive'] == 'REGISTRY_USER_S-1-5-21-2945379629-2233710143-2353048178-1000_CLASSES'
            assert kparams['key'] == "S-1-5-21-2945379629-2233710143-2353048178-1000_CLASSES\\Local Settings" \
                                     "\\Software\\Microsoft\\Windows\\CurrentVersion\\TrayNotify\\IconStreams"
            assert kparams['status'] == regkevt.status
            assert kparams['pid'] == regkevt.process_id
            assert kparams['tid'] == regkevt.thread_id
            assert kparams['value_type'] == 'REG_DWORD'
            assert kparams['value'] == '0x12'

    def test_parse_hive_delete_value(self, hive_parser, kevent_mock):
        regkevt = dd({"index": 0, "process_id": 1224, "status": 3221225524, "key_handle": 0,
                      "key_name": "\\Registry\\Machine\\Software\\Classes\\Applications\\Explorer.exe"
                                  "\\Drives\\C\\DefaultIcon", "thread_id": 2164, "initial_time": 24218563385})

        hive_parser.parse_hive(REG_DELETE_VALUE, regkevt)

        kparams = kevent_mock.params

        assert kparams['hive'] == 'REGISTRY_MACHINE_SOFTWARE'
        assert kparams['key'] == 'SOFTWARE\\Classes\\Applications\\Explorer.exe\\Drives\\C\\DefaultIcon'
        assert kparams['status'] == regkevt.status
        assert kparams['pid'] == regkevt.process_id
        assert kparams['tid'] == regkevt.thread_id

    def test_parse_hive_na(self, hive_parser, kevent_mock):
        regkevt = dd({"index": 2, "process_id": 896, "status": 0, "key_handle": 19446735964812642920,
                      "key_name": "ThreadingModel", "thread_id": 2916, "initial_time": 24219715376})
        hive_parser.parse_hive(REG_OPEN_KEY, regkevt)
        kparams = kevent_mock.params

        assert kparams['hive'] == NA
        assert kparams['key'] == '..\\ThreadingModel'





# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import json

import pika

from fibratus.errors import InvalidPayloadError
from fibratus.output.base import Output


class AmqpOutput(Output):

    def __init__(self, **kwargs):
        """Builds a new instance of the AMQP output adapter.

        Parameters
        ----------

        kwargs: dict
            AMQP configuration
        """
        Output.__init__(self)
        self._username = kwargs.pop('username', 'guest')
        self._password = kwargs.pop('password', 'guest')

        self._host = kwargs.pop('host', '127.0.0.1')
        self._port = kwargs.pop('port', 5672)
        self._vhost = kwargs.pop('vhost', '/')
        self._delivery_mode = kwargs.pop('delivery_mode', 1)

        credentials = pika.PlainCredentials(self._username, self._password)
        self._parameters = pika.ConnectionParameters(self._host,
                                                     self._port,
                                                     self._vhost,
                                                     credentials)

        self._exchange = kwargs.pop('exchange', None)
        self._routingkey = kwargs.pop('routingkey', None)

        self._connection = None
        self._channel = None

        self._basic_props = pika.BasicProperties(content_type='text/json',
                                                 delivery_mode=self._delivery_mode)

    def emit(self, body, **kwargs):
        if not self._connection:
            self._connection = pika.BlockingConnection(self._parameters)
            self._channel = self._connection.channel()
        # override the default exchange name
        # and the routing key used to send
        # the message to the AMQP broker
        self._routingkey = kwargs.pop('routingkey', self._routingkey)
        self._exchange = kwargs.pop('exchange', self._exchange)

        # the message body should be a dictionary
        if not isinstance(body, dict):
            raise InvalidPayloadError('invalid payload for AMQP message. '
                                      'dict expected but %s found'
                                      % type(body))
        body = json.dumps(body)
        self._channel.basic_publish(self._exchange,
                                    self._routingkey,
                                    body, self._basic_props)

    @property
    def username(self):
        return self._username

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def vhost(self):
        return self._vhost

    @property
    def exchange(self):
        return self._exchange

    @property
    def routingkey(self):
        return self._routingkey

    @property
    def delivery_mode(self):
        return self._delivery_mode



# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ctypes import windll, CDLL


ADVAPI = 0
KERNEL = 1
NT = 2
C = 3
USER = 4

__LIBS__ = {ADVAPI: windll.advapi32,
            KERNEL: windll.kernel32,
            NT: windll.ntdll,
            C: CDLL('msvcrt'),
            USER: windll.user32}


def declare(lib_name, function_name, args, restype):
    if lib_name in __LIBS__:
        lib = __LIBS__[lib_name]
        function = getattr(lib, function_name)
        if function:
            if len(args) > 0:
                function.argtypes = args
            if restype:
                function.restype = restype
        return function
    else:
        raise AttributeError('The library %s cannot be loaded' % lib_name)


# Copyright 2015/2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os
from unittest.mock import Mock, patch

import pytest
from apscheduler.schedulers.background import BackgroundScheduler

import fibratus.filament as fil
from fibratus.errors import FilamentError
from fibratus.filament import Filament
from fibratus.output.amqp import AmqpOutput
from fibratus.term import AnsiTerm
from tests.fixtures.filaments import test_filament


@pytest.fixture(scope='module')
def ansi_term_mock():
    return Mock(spec_set=AnsiTerm)


@pytest.fixture()
def filament(ansi_term_mock):
    fil.FILAMENTS_DIR = os.path.join(os.path.dirname(__file__), '..', 'fixtures\\filaments')
    f = Filament()
    f._ansi_term = ansi_term_mock
    f.scheduler = Mock(spec_set=BackgroundScheduler)
    return f


class TestFilament(object):

    def test_load_filament(self, filament):
        with patch('os.listdir', return_value=['top_connections.py', 'test_filament.py']):
            filament.load_filament('test_filament')
            assert filament._filament_module
            assert isinstance(filament._filament_module, type(test_filament))
            assert filament._filament_module.__doc__
            assert 'test_filament' in filament.name

            f = filament.filament_module
            assert hasattr(f, 'set_filter')
            assert hasattr(f, 'set_interval')
            assert hasattr(f, 'columns')
            assert hasattr(f, 'title')
            assert hasattr(f, 'sort_by')
            assert hasattr(f, 'limit')
            assert hasattr(f, 'add_row')
            assert hasattr(f, 'render_tabular')

            f.set_filter('CreateProcess', 'CreateThread')
            assert 'CreateProcess' in filament._filters
            assert 'CreateThread' in filament._filters

            f.set_interval(1)
            assert filament._interval == 1

            f.columns(['IP', 'Port', 'Process'])
            assert filament._cols == ['IP', 'Port', 'Process']

            f.limit(20)
            assert filament._limit == 20

            f.add_row(['192.168.1.3', '80', 'chrome.exe'])

            f.sort_by('Port')
            assert 'Port' in filament._sort_by
            assert filament._sort_desc

            f.title('Top outbound connections')

    def test_load_filament_not_found(self, filament):
        with patch('os.listdir', return_value=['top_connections.py', 'test_filament.py']):
            with pytest.raises(FilamentError):
                filament.load_filament('filament_notfound')

    def test_load_filament_nodoc(self, filament):
        with patch('os.listdir', return_value=['top_connections.py', 'test_filament_nodoc.py']):
            with pytest.raises(FilamentError) as e:
                filament.load_filament('test_filament_nodoc')
            assert "Please provide a short description for the filament" in str(e.value)

    def test_load_filament_no_on_next_kevent_method(self, filament):
        with patch('os.listdir', return_value=['top_connections.py', 'test_filament_no_on_next_kevent.py']):
            with pytest.raises(FilamentError) as e:
                filament.load_filament('test_filament_no_on_next_kevent')
            assert 'Missing required on_next_kevent method on filament' in str(e.value)

    def test_load_filament_wrong_on_next_kevent(self, filament):
        with patch('os.listdir', return_value=['test_filament_wrong_on_next_kevent.py']):
            with pytest.raises(FilamentError) as e:
                filament.load_filament('test_filament_wrong_on_next_kevent')
            assert 'one argument' in str(e.value)

    def test_close(self, filament):
        filament.close()

    def test_set_invalid_interval(self, filament):
        with patch('os.listdir', return_value=['top_connections.py', 'test_filament_invalid_interval.py']):
            with pytest.raises(FilamentError):
                filament.load_filament('test_filament_invalid_interval')

    def test_add_row_invalid_type(self, filament):
        with patch('os.listdir', return_value=['test_filament.py']):
            filament.load_filament('test_filament')
            f = filament.filament_module
            with pytest.raises(FilamentError):
                f.add_row({'ip': '192.168.4.31'})

    def test_sort_by_no_columns(self, filament):
        with patch('os.listdir', return_value=['test_filament.py']):
            filament.load_filament('test_filament')
            f = filament.filament_module
            with pytest.raises(FilamentError) as e:
                f.sort_by('Port')
            assert 'Expected at least 1 column but 0 found' in str(e.value)

    def test_sort_by_column_not_found(self, filament):
        with patch('os.listdir', return_value=['test_filament.py']):
            filament.load_filament('test_filament')
            f = filament.filament_module
            f.columns(['IP', 'Port', 'Process'])
            with pytest.raises(FilamentError) as e:
                f.sort_by('File')
            assert 'File column does not exist' in str(e.value)

    def test_limit_no_columns(self, filament):
        with patch('os.listdir', return_value=['test_filament.py']):
            filament.load_filament('test_filament')
            f = filament.filament_module
            with pytest.raises(FilamentError):
                f.limit(20)

    def test_limit_non_integer(self, filament):
        with patch('os.listdir', return_value=['test_filament.py']):
            filament.load_filament('test_filament')
            f = filament.filament_module
            f.columns(['IP', 'Port', 'Process'])
            with pytest.raises(FilamentError):
                f.limit('20')

    def test_do_output_accessors(self, filament):
        with patch('os.listdir', return_value=['test_filament.py']):
            filament.load_filament('test_filament')
            outputs = {'amqp': Mock(spec_set=AmqpOutput)}
            filament.do_output_accessors(outputs)
            assert getattr(filament.filament_module, 'amqp')

    def test_set_columns_not_list(self, filament):
        with patch('os.listdir', return_value=['test_filament.py']):
            filament.load_filament('test_filament')
            with pytest.raises(FilamentError):
                filament.filament_module.columns(('IP', 'Port'))

    def test_render_tabular(self, filament, ansi_term_mock):
        with patch('os.listdir', return_value=['test_filament.py']):
            filament.load_filament('test_filament')
            filament.filament_module.columns(['IP', 'Port'])
            filament.render_tabular()
            ansi_term_mock.setup_console.assert_called_once()
            ansi_term_mock.write_output.assert_called_once()



# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Performs the indexing of the kernel's event stream to
Elasticsearch on interval basis. When the scheduled
interval elapses, the list of documents aggregated
are indexed to Elasticsearch.
"""

from datetime import datetime
documents = []


def on_init():
    set_filter('CreateThread', 'CreateProcess', 'TerminateThread', 'TerminateProcess',
               'CreateFile', 'DeleteFile', 'WriteFile', 'RenameFile', 'Recv', 'Send',
               'Accept', 'Connect', 'Disconnect', 'LoadImage', 'UnloadImage',
               'RegCreateKey', 'RegDeleteKey', 'RegSetValue')
    set_interval(1)


def on_next_kevent(kevent):
    doco = {'image': kevent.thread.name,
            'thread': {
                'exe': kevent.thread.exe,
                'comm': kevent.thread.comm,
                'pid': kevent.thread.pid,
                'tid': kevent.tid,
                'ppid': kevent.thread.ppid},
            'category': kevent.category,
            'name': kevent.name,
            'ts': '%s %s' % (datetime.now().strftime('%m/%d/%Y'),
                             kevent.timestamp.strftime('%H:%M:%S.%f')),
            'cpuid': kevent.cpuid,
            'params': kevent.params}
    documents.append(doco)


def on_interval():
    if len(documents) > 0:
        elasticsearch.emit(documents)
        documents.clear()



# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest.mock import Mock

import pytest

from fibratus.common import DotD as dd
from fibratus.kevent import KEvent
from fibratus.kevent_types import SEND_SOCKET_TCPV4, SEND_SOCKET_UDPV4, RECV_SOCKET_TCPV4, RECV_SOCKET_UDPV4, \
    ACCEPT_SOCKET_TCPV4, CONNECT_SOCKET_TCPV4, DISCONNECT_SOCKET_TCPV4, RECONNECT_SOCKET_TCPV4
from fibratus.tcpip.tcpip import TcpIpParser


@pytest.fixture(scope='module')
def kevent_mock():
    return Mock(spec_set=KEvent)


@pytest.fixture(scope='module')
def tcpip_parser(kevent_mock):
    return TcpIpParser(kevent_mock)


class TestTcpIpParser():

    @pytest.mark.parametrize('ktcpip, l4_proto, l5_proto, kevent_type', [
        (dd({"saddr": "10.0.2.15", "sport": 49279, "daddr": "216.58.211.238",
             "dport": 443, "pid": 1848, "connid": 0, "size": 63, "seqnum": 0}), 'TCP', 'https', SEND_SOCKET_TCPV4),
        (dd({"saddr": "10.0.2.15", "sport": 49279, "daddr": "216.58.211.238",
             "dport": 53, "pid": 1848, "connid": 0, "size": 63, "seqnum": 0}), 'UDP', 'domain', SEND_SOCKET_UDPV4)])
    def test_parse_send(self, ktcpip, l4_proto, l5_proto,
                        kevent_type, tcpip_parser,
                        kevent_mock):

        tcpip_parser.parse_tcpip(kevent_type, ktcpip)
        kparams = kevent_mock.params

        assert kparams['pid'] == ktcpip.pid
        assert kparams['ip_src'] == ktcpip.saddr
        assert kparams['ip_dst'] == ktcpip.daddr
        assert kparams['sport'] == ktcpip.sport
        assert kparams['dport'] == ktcpip.dport
        assert kparams['packet_size'] == ktcpip.size
        assert kparams['l4_proto'] == l4_proto
        assert kparams['protocol'] == l5_proto

    @pytest.mark.parametrize('ktcpip, l4_proto, l5_proto, kevent_type', [
        (dd({'seqnum': 0, 'connid': 0, 'sport': 49720, 'daddr': '91.226.88.5',
             'saddr': '10.0.2.15', 'size': 266, 'pid': 1380, 'dport': 443}), 'TCP', 'https', RECV_SOCKET_TCPV4),
        (dd({"saddr": "10.0.2.15", "sport": 49279, "daddr": "216.58.211.238",
             "dport": 53, "pid": 1848, "connid": 0, "size": 63, "seqnum": 0}), 'UDP', 'domain', RECV_SOCKET_UDPV4)])
    def test_parse_recv(self, ktcpip, l4_proto, l5_proto,
                        kevent_type, tcpip_parser,
                        kevent_mock):

        tcpip_parser.parse_tcpip(kevent_type, ktcpip)
        kparams = kevent_mock.params

        assert kparams['pid'] == ktcpip.pid
        assert kparams['ip_src'] == ktcpip.saddr
        assert kparams['ip_dst'] == ktcpip.daddr
        assert kparams['sport'] == ktcpip.sport
        assert kparams['dport'] == ktcpip.dport
        assert kparams['packet_size'] == ktcpip.size
        assert kparams['l4_proto'] == l4_proto
        assert kparams['protocol'] == l5_proto

    def test_parse_recv_dport_na(self, tcpip_parser, kevent_mock):
        ktcpip = dd({'seqnum': 0, 'connid': 0, 'sport': 25, 'daddr': '91.226.88.5',
                    'saddr': '10.0.2.15', 'size': 266, 'pid': 1380, 'dport': 51234})
        tcpip_parser.parse_tcpip(RECV_SOCKET_TCPV4, ktcpip)
        kparams = kevent_mock.params
        assert kparams['protocol'] == 'smtp'

    def test_parse_accept(self, tcpip_parser, kevent_mock):
        ktcpip = dd({'connid': 0, 'sndwinscale': 0, 'rcvwinscale': 0, 'saddr': '10.0.2.15',
                     'sport': 22, 'rcvwin': 64240, 'tsopt': 0, 'pid': 1380, 'seqnum': 0,
                     'daddr': '216.58.211.206', 'size': 0, 'mss': 1460, 'dport': 49804, 'wsopt': 0, 'sackopt': 0})

        tcpip_parser.parse_tcpip(ACCEPT_SOCKET_TCPV4, ktcpip)
        kparams = kevent_mock.params

        assert kparams['pid'] == ktcpip.pid
        assert kparams['ip_src'] == ktcpip.saddr
        assert kparams['ip_dst'] == ktcpip.daddr
        assert kparams['sport'] == ktcpip.sport
        assert kparams['dport'] == ktcpip.dport
        assert kparams['rwin'] == ktcpip.rcvwin
        assert kparams['protocol'] == 'ssh'

    def test_parse_connect(self, tcpip_parser, kevent_mock):
        ktcpip = dd({'connid': 0, 'sndwinscale': 0, 'rcvwinscale': 0, 'saddr': '10.0.2.15',
                     'sport': 49804, 'rcvwin': 64240, 'tsopt': 0, 'pid': 1380, 'seqnum': 0,
                     'daddr': '216.58.211.206', 'size': 0, 'mss': 1460, 'dport': 443, 'wsopt': 0, 'sackopt': 0})

        tcpip_parser.parse_tcpip(CONNECT_SOCKET_TCPV4, ktcpip)
        kparams = kevent_mock.params

        assert kparams['pid'] == ktcpip.pid
        assert kparams['ip_src'] == ktcpip.saddr
        assert kparams['ip_dst'] == ktcpip.daddr
        assert kparams['sport'] == ktcpip.sport
        assert kparams['dport'] == ktcpip.dport
        assert kparams['rwin'] == ktcpip.rcvwin
        assert kparams['protocol'] == 'https'

    def test_parse_disconnect(self, tcpip_parser, kevent_mock):
        ktcpip = dd({'connid': 0, 'saddr': '10.0.2.15',
                     'sport': 49804, 'pid': 1380,
                     'daddr': '216.58.211.206', 'dport': 443,})

        tcpip_parser.parse_tcpip(DISCONNECT_SOCKET_TCPV4, ktcpip)
        kparams = kevent_mock.params

        assert kparams['pid'] == ktcpip.pid
        assert kparams['ip_src'] == ktcpip.saddr
        assert kparams['ip_dst'] == ktcpip.daddr
        assert kparams['sport'] == ktcpip.sport
        assert kparams['dport'] == ktcpip.dport

    def test_parse_reconnect(self, tcpip_parser, kevent_mock):
        ktcpip = dd({'connid': 0, 'saddr': '10.0.2.15',
                     'sport': 49804, 'pid': 1380,
                     'daddr': '216.58.211.206', 'dport': 443,})

        tcpip_parser.parse_tcpip(RECONNECT_SOCKET_TCPV4, ktcpip)
        kparams = kevent_mock.params

        assert kparams['pid'] == ktcpip.pid
        assert kparams['ip_src'] == ktcpip.saddr
        assert kparams['ip_dst'] == ktcpip.daddr
        assert kparams['sport'] == ktcpip.sport
        assert kparams['dport'] == ktcpip.dport








# Copyright 2017 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from fibratus.output.base import Output
import io
import os
import time
import json


class FsOutput(Output):
    """File system output.

    Implementation of the output which writes the stream of
    kernel events to a file.
    """

    def __init__(self, **kwargs):
        Output.__init__(self)
        self._path = kwargs.pop('path', None)
        self._fmt = kwargs.pop('format', 'json')
        self._mode = kwargs.pop('mode', 'a')

        filename = os.path.join(self._path,
                                '%s.fibra' % time.strftime('%x')
                                .replace('/', '-'))
        self.stream = io.open(filename, self._mode)

    def emit(self, body, **kwargs):
        if 'json' in self._fmt:
            self.stream.write(json.dumps(body) + '\n')

    @property
    def path(self):
        return self._path

    @property
    def format(self):
        return self._fmt

    @property
    def mode(self):
        return self._mode


# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import json
from _ctypes import byref

from fibratus.apidefs.sys import get_std_handle, STD_OUTPUT_HANDLE, write_console_unicode, c_ulong
from fibratus.output.base import Output

RENDER_FORMAT = '%s %s %s %s (%s) - %s %s'


class ConsoleOutput(Output):

    def __init__(self, **kwargs):
        Output.__init__(self)

        self._fmt = kwargs.pop('format', 'pretty')
        self._timestamp_pattern = kwargs.pop('timestamp_pattern', '%Y-%m-%d %H:%M:%S.%f')
        self._stdout_handle = get_std_handle(STD_OUTPUT_HANDLE)

        assert self._stdout_handle, 'could not acquire the standard output stream handle'

    def emit(self, kevent, **kwargs):
        """Renders the kevent to the standard output stream.

        Uses the default output format or JSON to render the
        kernel event to standard output stream.

        The default output format is as follows:

        id  timestamp  cpu  process  (process id) - kevent (parameters)
        --  ---------  ---  -------  -----------   ------- ------------

        Example:

        160 13:27:27.554 0 wmiprvse.exe (1012) - CloseFile (file=C:\\WINDOWS\\SYSTEM32\\RSAENH.DLL, tid=2668)

        Parameters
        ----------

        kevent: KEvent
            the information regarding the kernel event

        kwargs: dict
            console adapter configuration

        """
        if isinstance(kevent, dict):
            kevt = json.dumps(kevent)
        else:
            pid, proc = kevent.get_thread()
            if 'pretty' in self._fmt:
                kevt = RENDER_FORMAT % (kevent.kid,
                                        kevent.ts.time(),
                                        kevent.cpuid,
                                        proc,
                                        pid,
                                        kevent.name,
                                        self._format_params(kevent.params))
            else:
                kevt = json.dumps(dict(id=kevent.kid,
                                       timestamp=kevent.ts.strftime(self._timestamp_pattern),
                                       cpuid=kevent.cpuid,
                                       proc=proc,
                                       pid=pid,
                                       name=kevent.name,
                                       params=kevent.params))

        kevt += '\n'
        # write the output on the standard output stream
        write_console_unicode(self._stdout_handle, kevt,
                              len(kevt),
                              byref(c_ulong()),
                              None)

    def _format_params(self, kparams):
        """Transforms the kevent parameters.

        Apply the rendering format on the kevent payload
        to transform it into more convenient structure
        sorted by parameter keys.
        """
        fmt = ', '.join('%s=%s' % (k, kparams[k]) for k in sorted(kparams.keys())) \
            .replace('\"', '')
        return '(%s)' % fmt



# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import smtplib
from unittest.mock import patch, Mock

import pytest
from logbook import Logger

from fibratus.output.smtp import SmtpOutput


@pytest.fixture(scope='module')
def smtp_adapter():
    config = {
        'host': 'smtp.gmail.com',
        'from': 'fibratus@fibratus.io',
        'password': 'secret',
        'to': ['fibratus@fibratus.io', 'netmutatus@netmutatus.io']
    }
    return SmtpOutput(**config)


class TestSmtpOutput(object):

    def test_init(self, smtp_adapter):
        assert 'smtp.gmail.com' in smtp_adapter.host
        assert 'fibratus@fibratus.io' in smtp_adapter.sender
        assert set(['fibratus@fibratus.io', 'netmutatus@netmutatus.io']) == set(smtp_adapter.to)
        assert smtp_adapter.port == 587

    def test_emit(self, smtp_adapter):
        body = 'Anomalous network activity detected from notepad.exe process'
        with patch('smtplib.SMTP'):
            smtp_adapter.emit(body, subject='Anomalous network activity detected')
            assert smtp_adapter._smtp.ehlo.call_count == 2
            smtp_adapter._smtp.starttls.assert_called_once()
            smtp_adapter._smtp.login.assert_called_with('fibratus@fibratus.io', 'secret')
            message = 'From: fibratus@fibratus.io' \
                      'To: fibratus@fibratus.io, netmutatus@netmutatus.io' \
                      'Subject: Anomalous network activity detected' \
                      'Anomalous network activity detected from notepad.exe process'

            smtp_adapter._smtp.login.sendmail('fibratus@fibratus.io', ['fibratus@fibratus.io',
                                                                       'netmutatus@netmutatus.io'],
                                              message)
            smtp_adapter._smtp.quit.assert_called_once()

    def test_emit_invalid_credentials(self, smtp_adapter):
        body = 'Anomalous network activity detected from notpead.exe process'
        smtp_adapter.logger = Mock(spec_set=Logger)
        with patch('smtplib.SMTP'):
            smtp_adapter._smtp.login.side_effect = smtplib.SMTPAuthenticationError(534, 'Invalid smtp credentials')
            smtp_adapter.emit(body, subject='Anomalous network activity detected')
            smtp_adapter.logger.error.assert_called_with('Invalid SMTP credentials for '
                                                         'fibratus@fibratus.io account')
            smtp_adapter._smtp.quit.assert_called_once()
